/*
# #############################################################################
#  ATTENTION
# #############################################################################
#
#  This software is for guidance only. It is provided "as is".
#  It only aims at providing coding information to the customer in order for
#  them to save time. As a result no warranties, whether express, implied or
#  statutory, including, but not limited to implied warranties of
#  merchantability and fitness for a particular purpose apply to this software.
#  A·P·E Angewandte Physik und Elektronik GmbH shall not be held liable for any
#  direct, indirect or consequential damages with respect to any claims arising
#  from the content of such sourcecode and/or the use made by customers of the
#  coding information contained herein in connection with their products.
#
#  File Version: 1.0
#
#  Changelog:
#    1.0 - Initial Version
#
# #############################################################################
*/

#include "ape_device.hpp"
#include "stdafx.h"

APE_Device::APE_Device(std::string hostname, int port, std::string devicename = "")
{
	int iResult = 1;

	//Portnumber valid?
	if (port <= 0 || port > 65535)
	{
		return;
	}
	else
		remote_port = std::to_string(port);

	//Hostname or IP set?
	if (hostname.empty())
	{
		return;
	}
	else
		remote_host = hostname;
	
	//Custom devicename set?
	if (devicename.empty())
	{
		custom_name = devicename;
	}
	
	socket_timeout = 5.0;

	//Try to open connection
	do 
		iResult = open_connection();
	while (iResult > 0);
}

//-----------------------------------------------------------------------------
//Class destructor
//-----------------------------------------------------------------------------
APE_Device::~APE_Device()
{
	//Close connection if still active
	if (is_connected)
		disconnect();
}

bool APE_Device::connected()
{
	return is_connected;
}

std::string APE_Device::set_host(std::string hostname)
{
	//Hostname or IP set?
	if (hostname.size() <= 0)
	{
		return "";
	}
	else
		remote_host = hostname;

	return hostname;
}

unsigned int APE_Device::set_port(unsigned int port)
{
	//Portnumber valid?
	if (port <= 0 || port > 65535)
	{
		return 0;
	}
	else
		remote_port = std::to_string(port);

	return port;
}

float APE_Device::set_timeout(float timeout)
{
	//Portnumber valid?
	if (timeout <= 0)
	{
		return -1;
	}
	else
		socket_timeout = timeout;

	return socket_timeout;

}

int APE_Device::open_connection(void)
{
	int iResult;

	// Initialize Winsock
	iResult = WSAStartup(MAKEWORD(2, 2), &wsaData);
	if (iResult != 0) {
		printf("[Connect] Error. WSAStartup failed with error: %d\n", iResult);
		return 1;
	}
	ZeroMemory(&hints, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;

	// Resolve the server address and port
	iResult = getaddrinfo(remote_host.c_str(), remote_port.c_str(), &hints, &result);
	if (iResult != 0) {
		printf("[Connect] Error. Getaddrinfo failed with error: %d\n", iResult);
		WSACleanup();
		return 1;
	}

	// Attempt to connect to an address until one succeeds
	for (ptr = result; ptr != NULL; ptr = ptr->ai_next) {

		// Create a SOCKET for connecting to server
		ConnectSocket = socket(ptr->ai_family, ptr->ai_socktype,
			ptr->ai_protocol);
		if (ConnectSocket == INVALID_SOCKET) {
			printf("[Connect] Error. Socket failed with error: %ld\n", WSAGetLastError());
			WSACleanup();
			return 1;
		}

		// Connect to server.
		iResult = connect(ConnectSocket, ptr->ai_addr, (int)ptr->ai_addrlen);
		if (iResult == SOCKET_ERROR) {
			closesocket(ConnectSocket);
			ConnectSocket = INVALID_SOCKET;
			continue;
		}
		break;
	}
	freeaddrinfo(result);

	if (ConnectSocket == INVALID_SOCKET) {
		WSACleanup();
		return 1;
	}

	Sleep(500);

	is_connected = true;
	return 0;
}

bool APE_Device::disconnect(void)
{
	// Close the socket connection
	if (is_connected)
	{
		closesocket(ConnectSocket);
		WSACleanup();
		is_connected = false;
	}

	return is_connected;
}

int APE_Device::send_data(std::string command)
{
	int iResult;

	if (is_connected)
	{
		iResult = send(ConnectSocket, command.c_str(), command.size(), 0);
		if (iResult == SOCKET_ERROR) {
			printf("[Send] Send failed with error: %d\n", WSAGetLastError());
			return -1;
		}
		return iResult;
	}
	else
	{
		return -1;
	}

}

std::string APE_Device::recv_scpi()
{
	std::string buffer; 
	int header_len, data_len;

	//Get first character from socket, if it is not "#" the following data is no scpi block
	buffer = recv_data(1);

	//If first character of the data is not a #, this is not a valid scpi block
	if (buffer[0] != '#')
	{
		printf("[RECV_SCPI] Expected \"#\" but got %s", buffer);
		return "";
	}

	//read second character from socket, this is the length of the header as a decimal number
	buffer = recv_data(1);

	//Read header length and try to convert it to a number
	if (! (isdigit(buffer[0])))
	{
		printf("[RECV_SCPI] Expected decimal character but got %s", buffer);
		return "";
	}
	//Convert character to numeric value
	header_len = atoi(buffer.substr(0,1).c_str());

	if (header_len <= 0)
	{
		printf("[RECV_SCPI] Expected decimal character >= \"1\" but got %s", buffer);
		return "";
	}

	//Read number of bytes defined in the header_len  and try to convert it to a number
	buffer = recv_data(header_len);

	//Check if all received characters are numbers
	for (int i = 0; i < header_len; i++)
	{
		if (! isdigit(buffer[i]))
		{
			printf("[RECV_SCPI] Expected decimal character but got %s", buffer);
			return "";
		}
	}

	//Convert received data to numeric value
	data_len = atoi(buffer.c_str());

	//Data Length = 0 -> return empty string
	if (data_len <= 0)
		return "";

	//Read decoded amount of data and return as string object
	buffer = recv_data(data_len);
	return buffer;
}

std::string APE_Device::recv_data(int length, bool recv_scpi_block)
{
	int i, iResult, read_len;
	std::string outstr = "";
	
	if (! is_connected)
	{
		return "";
	}

	if (length == 0)
	return "";

	if ((length < 0) || (length > BUFFER_LEN))
		read_len = BUFFER_LEN;
	else
		read_len = length;

	//If an scpi block should be read, use extra function
	if (recv_scpi_block)
		return recv_scpi();


	//Read from socket until requested amount of data was read 
	//or (if one line was requested) "\n" was received
	do {
		iResult = recv(ConnectSocket, recvbuffer, read_len, 0);
		if (iResult > 0)
		{		
			for (i = 0; i < iResult; i++)
				outstr += recvbuffer[i];

			read_len -= iResult;
		}
	} while ((iResult > 0)&&(read_len > 0)&&(length > 0));
	int len = outstr.size();
	return outstr;
}


std::string APE_Device::query(std::string command, bool recv_scpi_block)
{
	std::string buffer;
	if (!is_connected)
	{
		return "";
	}

	if (command.size() <= 0)
	{
		return "";
	}

	send_data(command);

	buffer = recv_data(-1, recv_scpi_block);
	int len = buffer.size();
	return buffer;
}

std::string APE_Device::get_idn(void)
{
	return query("*IDN?\r\n");
}

int APE_Device::get_status(void)
{
	return atoi(query("*STB?\r\n").c_str());
}

int APE_Device::get_oper(void)
{
	return atoi(query("*OPER?\r\n").c_str());
}

